#!/usr/bin/env python

import os
from collections import namedtuple
import pickle
import gzip
import argparse
import numpy as np
import pandas as pd
from scipy import interpolate
from . import pl
from nuVeto.utils import calc_bins, centers


def hist_preach(infile):
    """ Builds histograms of P_reach based on MMC output text
    """
    Hist = namedtuple('Hist', 'counts edges')
    df = pd.read_csv(infile, delim_whitespace=True, header=None,
                     names='ei l ef'.split())
    # If the muon doesn't reach, MMC saves ef as -distance traveled
    df[df<0] = 0
    preach = []
    for (ei, l), efs in df.groupby(['ei', 'l']):
        bins = calc_bins(efs['ef'])
        histo = Hist(*np.histogram(efs['ef'], bins=bins, density=True))
        [preach.append((ei, l, ef, ew, val)) for ef,ew,val in zip(centers(histo.edges),
                                                                  np.ediff1d(histo.edges),
                                                                      histo.counts)]

    return np.asarray(preach)


def int_ef(preach, plight):
    """ integate p_reach*p_light over e_f to reduce dimensionality for interpolator
    """
    if isinstance(preach, str) and os.path.isfile(preach):
        try:
            preach = pickle.load(gzip.open(preach, 'rb'), encoding='latin1')
        except IOError:
            preach = hist_preach(preach)
    df = pd.DataFrame(preach, columns='ei l ef ew pdf'.split())
    intg = []
    for (ei, l), sdf in df.groupby(['ei', 'l']):
        intg.append((ei, l, np.sum(sdf['ew']*sdf['pdf']*plight(sdf['ef']))))

    return np.asarray(intg)


def interp(preach, plight):
    intg = int_ef(preach, plight)
    df = pd.DataFrame(intg, columns='ei l prpl'.split())
    df = df.pivot_table(index='ei', columns='l', values='prpl').fillna(0)
    return interpolate.RegularGridInterpolator((df.index,df.columns), df.values, bounds_error=False, fill_value=None)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Generate muon detection probability')
    parser.add_argument('mmc', metavar='MMC',
                    help='text file or pickled histogram containing MMC simulated data')
    parser.add_argument('--plight', default='pl_step_1000',
                        choices=[fn for fn in dir(pl) if fn.startswith('pl_')],
                        help='choice of a plight function to apply as defined in pl.py')
    parser.add_argument('--noconvolution', default=False, action='store_true',
                        help='Generate pdfs of preach from raw MMC output and save to pklz')
    parser.add_argument('-o', dest='output', default='mu.pkl',
                        help='output file. To be read in this needs to be in "nuVeto/data/prpl/"')

    args = parser.parse_args()
    if args.noconvolution:
        hpr = hist_preach(args.mmc)
        pickle.dump(hpr, gzip.open(args.output, 'wb'), protocol=-1)
    else:
        intp = interp(args.mmc, getattr(pl, args.plight))
        pickle.dump(intp, open(args.output, 'wb'), protocol=-1)
    print('Output pickled into {}'.format(args.output))



#!/usr/bin/env python

"""
Calculate the rate at which atmospheric neutrinos arrive at an underground
detector with no accompanying muons.
"""

# Copyright (c) 2014, Jakob van Santen <jvansanten@icecube.wisc.edu>
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import numpy

def overburden(cos_theta, depth=1950, elevation=2400):
	"""
	Overburden for a detector buried underneath a flat surface.
	
	:param cos_theta: cosine of zenith angle (in detector-centered coordinates)
	:param depth:     depth of detector (in meters below the surface)
	:param elevation: elevation of the surface above sea level
	
	:returns: an overburden [meters]
	"""
	# curvature radius of the surface (meters)
	r = 6371315 + elevation
	# this is secrety a translation in polar coordinates
	return (numpy.sqrt(2*r*depth + (cos_theta*(r-depth))**2 - depth**2) - (r-depth)*cos_theta)

def minimum_muon_energy(distance):
	"""
	Minimum muon energy required to survive the given thickness of ice with at
	least 1 TeV 50% of the time.
	
	:returns: minimum muon energy [GeV]
	"""
	# require that the muon have median energy 1 TeV
	b, c = 2.52151, 7.13834
	return 1e3*numpy.exp(1e-3*distance/(b) + 1e-8*(distance**2)/c)

def effective_costheta(costheta):
	"""
	Effective local atmospheric density correction from [Chirkin]_.
	
	.. [Chirkin] D. Chirkin. Fluxes of atmospheric leptons at 600-GeV - 60-TeV. 2004. http://arxiv.org/abs/hep-ph/0407078
	"""
	x = costheta
	p = [0.102573, -0.068287, 0.958633, 0.0407253, 0.817285]
	return numpy.sqrt((x**2 + p[0]**2 + p[1]*x**p[2] + p[3]*x**p[4])/(1 + p[0]**2 + p[1] + p[3]))

class fpe_context(object):
	"""
	Temporarily modify floating-point exception handling
	"""
	def __init__(self, **kwargs):
		self.new_kwargs = kwargs
	def __enter__(self):
		self.old_kwargs = numpy.seterr(**self.new_kwargs)
	def __exit__(self, *args):
		numpy.seterr(**self.old_kwargs)

elbert_params = {
	'elbert': {'a' : 14.5,
	           'p1': 0.757+1,
	           'p2': 5.25,
	           'p3': 1},
	'mu' :    {'a': 49.41898933142626,
	           'p1': 0.62585930096346309+1,
	           'p2': 4.9382653076505525,
	           'p3': 0.58038589096897897},
	'numu':   {'a': 79.918830537201231,
	           'p1': 0.46284463423687988+1,
	           'p2': 4.3799061061862314,
	           'p3': 0.31657956163506323},
	'nue':    {'a' : 0.80523974705259793,
	           'p1': 0.61852327151082975+1,
	           'p2': 9.7834540169712252,
	           'p3': 0.65126864602268075},
	'charm':  {'a' : 780.35285355003532/1e6,
	           'p1': -0.39555243513109928+1,
	           'p2': 7.3461490462825703,
	           'p3': 0.76688386541155051}
}

def elbert_yield(emin, primary_energy, primary_mass, cos_theta, kind='mu', differential=False):
	"""
	Evaluate the lepton yield of the given lepton family per shower for showers
	of a given type.
	
	:param emin:           lepton energy at which to evaluate
	:param primary_energy: total primary energy
	:param primary_mass:   primary atomic number
	:param cos_theta:      cosine of primary zenith angle
	:param kind:           family of parameterization (mu, numu, nue, or charm)
	:param differential:   if True, evaluate dN/dE at emin. Otherwise,
	                       evaluate N(>emin).
	
	:returns: either a differential yield [1/GeV] or cumulative yield [number]
	"""
	params = elbert_params[kind]
	a  = params['a']
	p1 = params['p1']
	p2 = params['p2']
	p3 = params['p3']
	
	En = primary_energy/primary_mass
	x = emin/En
	
	if kind == 'charm':
		decay_prob = 1.
	else:
		decay_prob = 1./(En*effective_costheta(cos_theta))
	
	with fpe_context(all='ignore'):
		icdf = numpy.where(x >= 1, 0., a*primary_mass*decay_prob*x**(-p1)*(1-x**p3)**p2)
		if differential:
			icdf *= (1./En)*numpy.where(x >= 1, 0., (p1/x + p2*p3*x**(p3-1)/(1-x**p3)))
	
	return icdf

class ParticleType(object):
	PPlus       =   14
	He4Nucleus  =  402
	N14Nucleus  = 1407
	Al27Nucleus = 2713
	Fe56Nucleus = 5626

def gaisser_flux(energy, ptype):
	"""
	Evaluate the [Gaisser]_ H3a parameterization of the cosmic ray flux.
	
	:param energy: total energy of primary nucleus [GeV]
	:param ptype: particle type code 
	
	:returns: particle flux [particles/(GeV m^2 sr s)]
	
	.. [Gaisser] T. K. Gaisser. Spectrum of cosmic-ray nucleons, kaon production, and the atmospheric muon charge ratio. Astroparticle Physics, 35(12):801--806, 2012. ISSN 0927-6505. doi: 10.1016/j.astropartphys.2012.02.010.
	"""
	if ptype < 100:
		z = 1
	else:
		z = ptype % 100
	
	codes = sorted([v for v in list(ParticleType.__dict__.values()) if isinstance(v, int)])
	idx = codes.index(ptype)
	
	# normalizations for each element
	norm = [
		[7860., 3550., 2200., 1430., 2120.],
		[20]*2 + [13.4]*3,
		[1.7]*2 + [1.14]*3,
	]
	# spectral indices
	gamma = [
		[2.66, 2.58, 2.63, 2.67, 2.63],
		[2.4]*5,
		[2.4]*5
	]
	# cutoff rigitity
	rigidity = [
		4e6, 30e6, 2e9
	]
	
	return sum(n[idx]*energy**(-g[idx])*numpy.exp(-energy/(r*z)) for n, g, r in zip(norm, gamma, rigidity))

def logspace(start, stop, num):
	"""
	A version of numpy.logspace that takes array arguments
	"""
	# Add a trailing dimension to the inputs
	start = numpy.asarray(start)[...,None]
	stop = numpy.asarray(stop)[...,None]
	num = int(num)
	step = (stop-start)/float(num-1)
	y = (numpy.core.numeric.arange(0, num)*step + start).squeeze()
	return 10**y

def response_function(enu, emu, cos_theta, kind='numu'):
	"""
	Evaluate the response function (contributions from each chunk of the cosmic
	ray flux) for leptons of the given energy and zenith angle
	
	:param enu:       lepton energy
	:param emu:       minimum muon energy needed for veto [GeV at surface]
	:param cos_theta: cosine of zenith angle
	:param kind:      lepton type for which to evaluate the passing rate:
	                  "numu" for muon neutrinos from pion/kaon decay, "nue" for
	                  electron neutrinos from kaon decay, or "charm" for either
	                  flavor from charmed meson decay
	
	:returns: a tuple (response, muonyield)
	"""
	# make everything an array
	enu, emu, cos_theta = list(map(numpy.asarray, (enu, emu, cos_theta)))
	shape = numpy.broadcast(enu, emu, cos_theta).shape
	# contributions to the differential neutrino flux from chunks of the
	# primary spectrum for each element
	contrib = numpy.zeros(shape+(5, 100))
	# mean integral muon yield from same chunks
	muyield = numpy.zeros(shape+(5, 100))
	energy_per_nucleon = logspace(numpy.log10(enu), numpy.log10(enu)+3, 101)
	ptypes = [getattr(ParticleType, pt) for pt in ('PPlus', 'He4Nucleus', 'N14Nucleus', 'Al27Nucleus', 'Fe56Nucleus')]
	A = [[pt/100, 1][pt == ParticleType.PPlus] for pt in ptypes]
	for i, (ptype, a) in enumerate(zip(ptypes, A)):
		# primary energies that contribute to the neutrino flux at given energy
		penergy = a*energy_per_nucleon
		# width of energy bins
		de = numpy.diff(penergy)
		# center of energy bins
		pe = penergy[...,:-1] + de/2.
		# hobo-integrate the flux
		weights = gaisser_flux(pe, int(ptype))*de
		# neutrino yield from each chunk of the primary flux
		ey = elbert_yield(enu[...,None], pe, a, cos_theta[...,None], kind=kind, differential=True)
		contrib[...,i,:] = ey*weights
		muyield[...,i,:] = elbert_yield(emu[...,None], pe, a, cos_theta[...,None], kind='mu', differential=False)
	
	return contrib, muyield, energy_per_nucleon[...,:-1] + numpy.diff(energy_per_nucleon)/2.

def uncorrelated_passing_rate(enu, emu, cos_theta, kind='numu'):
	"""
	Calculate the probability that neutrinos of the given energy and type will
	be accompanied by at least one muon from an unrelated branch of the shower.
	
	:param enu:       neutrino energy
	:param emu:       minimum muon energy needed for veto [GeV at surface]
	:param cos_theta: cosine of zenith angle
	:param kind:      neutrino type for which to evaluate the passing rate:
	                  "numu" for muon neutrinos from pion/kaon decay, "nue" for
	                  electron neutrinos from kaon decay, or "charm" for either
	                  flavor from charmed meson decay
	"""
	# get contributions to the differential neutrino flux from chunks of
	# the cosmic ray spectrum in each element
	contrib, muyield = response_function(enu, emu, cos_theta, kind)[:2]
	# normalize contributions over all primaries
	contrib /= contrib.sum(axis=(-1, -2), keepdims=True)
	# weight contributions by probability of having zero muons. if that 
	# probability is always 1, then this returns 1 by construction
	return (numpy.exp(-muyield)*contrib).sum(axis=(-1, -2))

def analytic_numu_flux(enu, cos_theta, emu=None):
	"""
	Calculate the differential flux of muon-neutrinos from pion and kaon decay
	in air showers, optionally restricting the solution to the part of the
	decay phase space where the lab-frame energy of the muon is above the given
	threshold [Schoenert]_. See [Lipari]_ for a detailed derivation of the
	cascade-equation solution.
	
	:param enu:       neutrino energy at which to evaluate [GeV]
	:param cos_theta: cosine of zenith angle in detector-centered coordinates
	:param emu:       if not None, calculate only the fraction of the flux
	                  where the partner muon has at least the given energy.
	                  Otherwise, calculate the total flux.
	
	:returns: the neutrino flux as a fraction of the primary nucleon flux at
	          evaluated at the given energy.
	
	.. [Schoenert] S. Schoenert, T. K. Gaisser, E. Resconi, and O. Schulz. Vetoing atmospheric neutrinos in a high energy neutrino telescope. Phys. Rev. D, 79:043009, Feb 2009. doi: 10.1103/PhysRevD.79.043009.
	.. [Lipari] P. Lipari. Lepton spectra in the earth's atmosphere. Astroparticle Physics, 1 (2):195--227, 1993. ISSN 0927-6505. doi: 10.1016/0927-6505(93)90022-6.
	"""
	
	# Spectral index of the integral nucleon flux at the top of the atmosphere
	GAMMA = 1.7
	# Spectrum-weighted moments for nucleon and meson production
	Z_NN, Z_NPI, Z_NK, Z_PIPI,	Z_KK = 0.298, 0.079, 0.0118,  0.271, 0.223
	# Critical energies for pions and kaons above which re-interaction is more
	# likely than decay in flight
	EPS_PI, EPS_K = 115.,  850.
	# Useful constants
	R_PI, R_K, ALAM_N, ALAM_PI, ALAM_K = .5731, 0.0458,	  120.,	   160.,   180.

	F = (GAMMA + 2.)/(GAMMA + 1.)
	B_PI = F * (ALAM_PI - ALAM_N) / (ALAM_PI * numpy.log(ALAM_PI/ALAM_N)*(1.-R_PI))
	B_K	 = F * (ALAM_K	- ALAM_N) / ( ALAM_K * numpy.log(ALAM_K /ALAM_N)*(1.-R_K))
	
	if emu is not None:
		z = 1 + emu/enu
		zpimin = 1./(1.-R_PI)
		zkmin = 1./(1.-R_K)
		zzpi = numpy.where(z >= zpimin, z, zpimin)
		zzk	 = numpy.where(z >= zkmin, z, zkmin)
		A_PI = Z_NPI/((1.-R_PI)*(GAMMA+1)*zzpi**(GAMMA+1.))
		A_K =  Z_NK/((1.-R_K)*(GAMMA+1)*zzk**(GAMMA+1.))
		
		B_PI = zzpi*B_PI*(1.-R_PI)
		B_K = zzk*B_K*(1.-R_K)
	else:
		A_PI = (Z_NPI*((1.-R_PI)**(GAMMA+1.))/(GAMMA+1.))/(1.-R_PI)
		A_K =  (Z_NK *((1.-R_K )**(GAMMA+1.))/(GAMMA+1.))/(1.-R_K)

	CS = effective_costheta(cos_theta)
	return (A_PI / ( 1. + B_PI*CS*enu/EPS_PI ) + 0.635 * A_K / ( 1. + B_K*CS*enu/EPS_K ))

def correlated_passing_rate(enu, emu, cos_theta):
	"""
	Calculate the probability that muon neutrinos of the given energy will be
	accompanied by a muon from the same decay vertex.
	
	:param enu:       neutrino energy
	:param emu:       minimum muon energy needed for veto [GeV at surface]
	:param cos_theta: cosine of zenith angle
	"""
	flux = analytic_numu_flux(enu, cos_theta, None)
	sflux = analytic_numu_flux(enu, cos_theta, emu)
	return (flux-sflux)/flux

# Cynthia Brewer's human-friendly colors http://colorbrewer2.org
colors = [(0.8941176470588236, 0.10196078431372549, 0.10980392156862745),
          (0.21568627450980393, 0.49411764705882355, 0.7215686274509804),
          (0.30196078431372547, 0.6862745098039216, 0.2901960784313726),
          (0.596078431372549, 0.3058823529411765, 0.6392156862745098),
          (1.0, 0.4980392156862745, 0.0),
          (1.0, 1.0, 0.2),
          (0.6509803921568628, 0.33725490196078434, 0.1568627450980392),
          (0.9686274509803922, 0.5058823529411764, 0.7490196078431373),
          (0.6, 0.6, 0.6)]

def plot_passing_rate(depth):
	import pylab
	ax = pylab.gca()
	enu = numpy.logspace(3, 7, 101)
	
	for zenith, color in zip(reversed((0, 41.4, 69.5, 81.4)), colors):
		cos_theta = numpy.cos(numpy.radians(zenith))
		emu = minimum_muon_energy(overburden(cos_theta, depth))
		correlated = correlated_passing_rate(enu, emu, cos_theta)
		uncorr_numu = uncorrelated_passing_rate(enu, emu, cos_theta, kind='numu')
		uncorr_nue = uncorrelated_passing_rate(enu, emu, cos_theta, kind='nue')
		
		pylab.plot(enu, correlated, color=color, label='%d' % zenith)
		pylab.plot(enu, correlated*uncorr_numu, color=color, ls=':')
		pylab.plot(enu, uncorr_nue, color=color, ls='--')
	
	pylab.semilogx()
	pylab.ylim((0, 1))
	pylab.ylabel('Passing fraction')
	pylab.xlabel('Neutrino energy [GeV]')
	olegend = pylab.legend(loc='lower left', title='Zenith [deg]')
	pylab.legend(ax.lines[:3], (r'$\nu_{\mu}$ (Correlated)', r'$\nu_{\mu}$ (Total)', r'$\nu_{e}$'),
	     loc='center right', prop=dict(size='small'))
	ax.add_artist(olegend)
	pylab.title('Atmospheric neutrino self-veto at %d m depth' % (opts.depth))
	
	pylab.show()

def format_energy(fmt, energy):
	places = int(numpy.log10(energy)/3)*3
	if places == 1:
		unit = 'GeV'
	elif places == 3:
		unit = 'TeV'
	elif places == 6:
		unit = 'PeV'
	elif places == 9:
		unit = 'EeV'
	return (fmt % (energy/10**(places))) + ' ' + unit

def plot_response_function(enu, depth, cos_theta, kind):
	
	emu = minimum_muon_energy(overburden(cos_theta, depth))
	response, muyield, energy_per_nucleon = response_function(enu, emu, cos_theta, kind)
	
	import pylab
	from matplotlib.gridspec import GridSpec
	from matplotlib.ticker import NullFormatter
	
	fig = pylab.figure()
	grid = GridSpec(3,1, height_ratios =[3,1,1], hspace=0)
	
	ax = pylab.subplot(grid[0])
	elements = ['H', 'He', 'C, N, O', 'Mg, Si, Al', 'Fe']
	for i, color, label in zip(list(range(5)), colors, elements):
		pylab.plot(energy_per_nucleon, response[i,:], color=color, lw=0.5, label=label)
		pylab.plot(energy_per_nucleon, numpy.exp(-muyield[i,:])*response[i,:], color=color, ls='--', lw=0.5)
	
	ax.plot(energy_per_nucleon, response.sum(axis=0), color='k', lw=1, label='Total')
	ax.plot(energy_per_nucleon, (numpy.exp(-muyield)*response).sum(axis=-2), color='k', ls='--', lw=1, label='(no muons)')
		
	pylab.loglog()
	ax.set_ylabel('Partial flux $[GeV^{-1}\, m^{-2}\, sr^{-1}\, s^{-1}]$')
	ax.legend(loc='best', prop=dict(size='small'), title='Flux contributions')
	passrate = (numpy.exp(-muyield)*response).sum(axis=(-2, -1))/response.sum(axis=(-2, -1))
	ax.set_title('%s %s from %d deg at %d m depth: %d%% passing rate' %
	    (format_energy('%d', enu), kind, numpy.round(numpy.degrees(numpy.arccos(cos_theta))), depth, passrate*100))
	
	logmax = numpy.ceil(numpy.log10(response.max()))
	ax.set_ylim(10**(logmax-8), 10**(logmax))
	ax.xaxis.set_major_formatter(NullFormatter())
	ax.yaxis.set_ticks(ax.yaxis.get_ticklocs()[2:-1])
	
	ax = pylab.subplot(grid[1])
	for i, color, label in zip(list(range(5)), colors, elements):
		pylab.plot(energy_per_nucleon, muyield[i,:], color=color, lw=0.5, label=label)
	pylab.loglog()
	logmax = numpy.ceil(numpy.log10(muyield.max()))
	ax.set_ylim(10**(logmax-4), 10**(logmax))
	ax.xaxis.set_major_formatter(NullFormatter())
	ax.set_ylabel(r'$\left<N_{\mu} > 1\, TeV \right>$')
	ax.yaxis.set_ticks(ax.yaxis.get_ticklocs()[2:-1])
	
	ax = pylab.subplot(grid[2])
	for i, color, label in zip(list(range(5)), colors, elements):
		pylab.plot(energy_per_nucleon, numpy.exp(-muyield[i,:]), color=color, lw=0.5, label=label)
	pylab.semilogx()
	ax.set_ylabel(r'$P(N_{\mu} = 0)$')
	
	pylab.xlabel('Energy per nucleon [GeV]')
	
	for ax in fig.axes:
		ax.grid(color='grey', ls='-', lw=0.1)
		ax.yaxis.set_label_coords(-.07, .5)
	
	help="""\
	The solid lines in the upper panel of the displayed plot show
	the contributions of cosmic ray primaries to the differential neutrino flux at
	%s and %d degrees. The center panel shows the mean number of muons from each
	shower that reach a vertical depth of %d meters with more than 1 TeV of kinetic energy.
	The bottom panel shows the Poisson probability of observing 0 muons at depth
	given the mean in the center panel. These probabilities are multiplied with the flux
	contributions to obtain the effective contributions that can be observed if
	accompanying muons are rejected, shown as dashed lines in the upper panel. The
	passing rate is the ratio of observable to the total flux, which is the ratio
	of the integrals of the solid and dashed black curves (%d%%).
	""" % (format_energy('%d', enu), numpy.round(numpy.degrees(numpy.arccos(cos_theta))), depth, passrate*100)

	import textwrap
	for line in textwrap.wrap(textwrap.dedent(help), 78):
		print(line)
	
	pylab.show()

if __name__ == "__main__":
	
	from optparse import OptionParser, OptionGroup
	import sys
	parser = OptionParser(usage="%prog [OPTIONS]", description=__doc__)
	parser.add_option("--flavor", choices=('numu', 'nue'), default='numu',
	    help="[default: %default]")
	parser.add_option("--charm", action="store_true", default=False,
	    help="Evaluate self-veto probability for neutrinos from charm rather than conventional decay")
	parser.add_option("--depth", type=float, default=1950.,
	    help="Vertical depth at which to calculate veto probability [default: %default]")
	group = OptionGroup(parser, "Interactive plotting options", "The following options can be used to display an interactive plot instead of writing a table. This requires matplotlib.")
	group.add_option("--plot", choices=('None', 'passingrate', 'response'),
	    help="Choices: 'passingrate' for a plot of veto passing rates, 'response' for a plot of the neutrino response function and associated muon yields")
	group.add_option("--energy", type=float, default=1e4,
	    help="Evaluate response function for this neutrino energy [%default GeV]")
	group.add_option("--zenith", type=float, default=60,
	    help="Evaluate response for this zenith angle [%default degrees]")
	parser.add_option_group(group)
	opts, args = parser.parse_args()
	
	if opts.plot == 'passingrate':
		plot_passing_rate(opts.depth)
		sys.exit(0)
	elif opts.plot == 'response':

		plot_response_function(opts.energy, opts.depth,
		    numpy.cos(numpy.radians(opts.zenith)), [opts.flavor, 'charm'][opts.charm])
		sys.exit(0)
	
	# Calculate passing rate on a grid of energies and zenith angles
	enu = numpy.logspace(3, 7, 101)
	cos_theta = numpy.arange(0, 1, .05)+.05
	enu, cos_theta = numpy.meshgrid(enu, cos_theta)
	emu = minimum_muon_energy(overburden(cos_theta, opts.depth))
	
	if opts.flavor == 'numu':
		passrate = correlated_passing_rate(enu, emu, cos_theta)
	else:
		passrate = numpy.ones(enu.shape)
	
	if opts.charm:
		kind = 'charm'
	else:
		kind = opts.flavor
	passrate *= uncorrelated_passing_rate(enu, emu, cos_theta, kind=kind)
	
	data = numpy.vstack(list(map(numpy.ndarray.flatten, (cos_theta, overburden(cos_theta, opts.depth), emu, enu, passrate)))).T
	fields = ['cos_theta', 'overburden', 'emu_min', 'energy', 'passrate']
	
	numpy.savetxt(sys.stdout, data, fmt='%12.3e', header='\t'.join(['%12s' % s for s in fields])[1:])



#!/usr/bin/env python
"""
Generates initial muon energies and overburdens for input to MMC

Run as
./reaching.py | ./ammc -raw -user -sdec -lpm -bs=1 -ph=3 -bb=2 -sh=2 -scat -medi=ice -frho -f -r -vcut=1e-3 -cont > reaching.txt &

From https://icecube.wisc.edu/~dima/work/MUONPR/MUON/WEBMMC/HISTORY.TXT
    "Several new cross sections are implemented: Andreev, Bezrukov, and Bugaev
    (abb) parameterization of the bremsstrahlung cross section can now be selected
    with -bs=2. The previously available kkp (Kelner, Kokoulin, and Petrukhin) can
    be chosen with -bs=1.

    The options -allm and -phnu are now obsolete. A new option -ph=[1-4] chooses
    Bezrukov and Bugaev photonuclear cross section parameterization (1), Bezrukov
    and Bugaev with hard component from Bugaev and Shlepin (2), ALLM (3), and
    Butkevich and Mikheyev (4) parameterization. Another option -bb (similar to
    the former -phnu) chooses one of the parameterizations of the soft component
    in the BB cross section (for -ph=1-2, values of -bb=3/4 correspond to BB/ZEUS
    photon-nucleon cross section formula) or version of the ALLM parameterization
    (for -ph=3, values of -bb=1/2 choose the ALLM91/ALLM97 parameter sets).

    New option -sh=[1-2] is introduced. It allows to choose the nuclear structure
    function calculation algorithm for -ph=[3-4] according to Dutta-Seckel or
    Butkevich-Mikheyev papers.

    Hadronic channels of the tau decay are now described as two-body decays into a
    tau neutrino and a pion or a rho-770, a1-1260, or rho-1450 resonance."
"""
import numpy as np

nsamples = 1000
xvec = np.logspace(np.log10(1./1000), np.log10(1./(2e5)), 100)
emuis = np.logspace(2, 8, 100)

for distance in 1/xvec:
    for emui in emuis:
        # roughly assume 1GeV/4m minimum energy loss
        if emui < distance/4.:
            continue
        for i in range(nsamples):
            print(emui, distance)



import numpy as np
from scipy import stats, interpolate


def sigmoid(emu, center, scale):
    return stats.norm.cdf(emu, loc=center, scale=scale)


def modified_sigmoid(emu, k0, x0, k1, x1, c):
    x = emu
    return 1 / (1 + np.exp(-k0*(np.log10(x)-x0)))+c*np.exp(-(np.log10(x)-x1)**2/k1)


def pl_step_1000(emu):
    """ heaviside at 1 TeV
    """
    return sigmoid(emu, 1e3, 1e-6)


def pl_step_750(emu):
    """ heaviside at 750 GeV
    """
    return sigmoid(emu, 0.75e3, 1e-6)


def pl_sigmoid_750_100(emu):
    """ sigmoid centered at 750 GeV
    """
    return sigmoid(emu, 750, 100)


def pl_sigmoid_750_250(emu):
    """ sigmoid centered at 750 GeV
    """
    return sigmoid(emu, 750, 250)


def pl_hese(emu):
    return modified_sigmoid(emu, 2.48135679,  3.57243996,  2.05736124,  2.60328639,  0.45801598)*sigmoid(emu, 1., 1e-6)


def pl_noearlymu(emu):
    nstrarr = np.fromfile(
        '/Users/tianlu/Projects/icecube/studies/hydrangea/lh/n_str_prob_pspl_full_flat_caus_30_combined.txt', sep=' ')
    nstrarr = nstrarr.reshape(2, nstrarr.shape[0]/2)
    # take P(Nstr|Ee) likelihood from chaack
    nstr = lambda ee: 10**interpolate.interp1d(
        np.log10(nstrarr[0]), np.log10(nstrarr[1]),
        kind='linear', bounds_error=False, fill_value='extrapolate')(np.log10(ee))
    return 1-nstr(emu)



import numpy as np
from functools import partial
from nuVeto.external import selfveto


def joint_passing_rate(enu, emu, cth, kind='numu'):
    return selfveto.correlated_passing_rate(enu, emu, cth)*selfveto.uncorrelated_passing_rate(enu, emu, cth, kind=kind)


def passrates(kind='conv nu_mu'):
    find_pf = {'conv nu_mu':partial(joint_passing_rate, kind='numu'),
               'pr nu_mu':partial(joint_passing_rate, kind='charm'),
               'conv nu_e':partial(selfveto.uncorrelated_passing_rate, kind='nue'),
               'pr nu_e':partial(selfveto.uncorrelated_passing_rate, kind='charm')}
    return find_pf[kind.replace('bar', '')]


def corr(kind='conv nu_mu'):
    find_pf = {'conv nu_mu':selfveto.correlated_passing_rate,
               'pr nu_mu':selfveto.correlated_passing_rate}
    return find_pf[kind.replace('bar', '')]


def uncorr(kind='conv nu_mu'):    
    find_pf = {'conv nu_mu':partial(selfveto.uncorrelated_passing_rate, kind='numu'),
               'pr nu_mu':partial(selfveto.uncorrelated_passing_rate, kind='charm'),
               'conv nu_e':partial(selfveto.uncorrelated_passing_rate, kind='nue'),
               'pr nu_e':partial(selfveto.uncorrelated_passing_rate, kind='charm')}
    return find_pf[kind.replace('bar', '')]
    



"""nuVeto calculation [ref.]

This module computes the probability that an atmospheric neutrino will be
accompanied by a sibling muon produced in the same cosmic ray airshower at a
given depth.


"""

from functools import lru_cache
from pkg_resources import resource_filename
import numpy as np
import scipy.integrate as integrate
import scipy.interpolate as interpolate
import MCEq.core
import MCEq.data

from MCEq.core import MCEqRun
import crflux.models as pm
import mceq_config as config
from .utils import Units, ParticleProperties, MuonProb, Geometry, amu, centers
from .uncertainties import BARR, barr_unc

class nuVeto(object):
    """Class for computing the neutrino passing fraction i.e. (1-(Veto probability))"""
    def __init__(self, costh,
                 pmodel=(pm.HillasGaisser2012, 'H3a'),
                 hadr='SIBYLL2.3c', barr_mods=(), depth=1950*Units.m,
                 density=('CORSIKA', ('SouthPole', 'December')),
                 debug_level=1):
        """Initializes the nuVeto object for a particular costheta, CR Flux,
        hadronic model, barr parameters, and depth

        Note:
            A separate MCEq instance needs to be created for each
            combination of __init__'s arguments. To access pmodel and hadr,
            use mceq.pm_params and mceq.yields_params
        Args:
            costh (float): Cos(theta), the cosine of the neutrino zenith at the detector
            pmodel (tuple(CR model class, arguments)): CR Flux
            hadr (str): hadronic interaction model
            barr_mods: barr parameters
            depth (float): the depth at which the veto probability is computed below the ice
        """
        self.costh = costh
        self.pmodel = pmodel
        self.geom = Geometry(depth)
        theta = np.degrees(np.arccos(self.geom.cos_theta_eff(self.costh)))

        config.debug_level = debug_level
        # config.enable_em = False
        config.enable_muon_energy_loss = False
        config.return_as = 'total energy'
        config.adv_set['allowed_projectiles'] = [2212, 2112,
                                                 211, -211,
                                                 321, -321,
                                                 130,
                                                 -2212, -2112]#, 11, 22]
        config.ctau = 2.5
        self.mceq = MCEqRun(
            # provide the string of the interaction model
            interaction_model=hadr,
            # primary cosmic ray flux model
            # support a tuple (primary model class (not instance!), arguments)
            primary_model=pmodel,
            # zenith angle \theta in degrees, measured positively from vertical direction at surface
            theta_deg=theta,
            # atmospheric density model
            density_model=density)


        if len(barr_mods) > 0:
            for barr_mod in barr_mods:
                # Modify proton-air -> mod[0]
                self.mceq.set_mod_pprod(2212, BARR[barr_mod[0]].pdg, barr_unc, barr_mod)
            # Populate the modifications to the matrices by re-filling the interaction matrix
            self.mceq.regenerate_matrices(skip_decay_matrix=True)

        X_vec = np.logspace(np.log10(2e-3),
                            np.log10(self.mceq.density_model.max_X), 12)
        self.dX_vec = np.diff(X_vec)
        self.X_vec = 10**centers(np.log10(X_vec))


    @staticmethod
    def categ_to_mothers(categ, daughter):
        """Get the parents for this category"""
        rcharge = '-' if 'bar' in daughter else '+'
        lcharge = '+' if 'bar' in daughter else '-'
        rbar = 'bar' if 'bar' in daughter else ''
        lbar = '' if 'bar' in daughter else 'bar'
        if categ == 'conv':
            mothers = ['pi'+rcharge, 'K'+rcharge, 'K_L0']
            if 'nu_tau' in daughter:
                mothers = []
            elif 'nu_e' in daughter:
                mothers.extend(['K_S0', 'mu'+rcharge])
            elif 'nu_mu' in daughter:
                mothers.extend(['mu'+lcharge])
        elif categ == 'pr':
            if 'nu_tau' in daughter:
                mothers = ['D'+rcharge, 'D_s'+rcharge]
            else:
                mothers = ['D'+rcharge, 'D_s'+rcharge, 'D'+rbar+'0']#, 'Lambda'+lbar+'0']#, 'Lambda_c'+rcharge]
        elif categ == 'total':
            mothers = nuVeto.categ_to_mothers('conv', daughter)+nuVeto.categ_to_mothers('pr', daughter)
        else:
            mothers = [categ,]
        return mothers


    @staticmethod
    def esamp(enu, accuracy):
        """ returns the sampling of parent energies for a given enu
        """
        # TODO: replace 1e8 with MMC-prpl interpolated bounds
        return np.logspace(np.log10(enu),
                           np.log10(enu+1e8), int(1000*accuracy))


    @staticmethod
    def projectiles():
        """Get allowed pimaries"""
        pdg_ids = config.adv_set['allowed_projectiles']
        namer = ParticleProperties.modtab.pdg2modname
        allowed = []
        for pdg_id in pdg_ids:
            allowed.append(namer[pdg_id])
            try:
                allowed.append(namer[-pdg_id])
            except KeyError:
                continue
        return allowed


    @staticmethod
    def nbody(fpath, esamp, enu, fn, l_ice):
        with np.load(fpath) as dfile:
            xmus = centers(dfile['xedges'])
            xnus = np.concatenate([xmus, [1]])
            vals = np.nan_to_num(dfile['histograms'])

            ddec = interpolate.RegularGridInterpolator((xnus, xmus), vals,
                                                       bounds_error=False, fill_value=None)
            emu_mat = xmus[:,None]*esamp[None,:]*Units.GeV
            pmu_mat = ddec(np.stack(np.meshgrid(enu/esamp, xmus), axis=-1))
            reaching = 1-np.sum(pmu_mat*fn.prpl(np.stack([emu_mat, np.ones(emu_mat.shape)*l_ice], axis=-1)), axis=0)
            reaching[reaching < 0.] = 0.
            return reaching


    @staticmethod
    @lru_cache(2**12)
    def psib(l_ice, mother, enu, accuracy, prpl):
        """ returns the suppression factor due to the sibling muon
        """
        esamp = nuVeto.esamp(enu, accuracy)
        fn = MuonProb(prpl)
        if mother in ['D0', 'D0-bar']:
            reaching = nuVeto.nbody(
                resource_filename(
                'nuVeto','data/decay_distributions/D0_numu.npz'),
                esamp, enu, fn, l_ice)
        elif mother in ['D+', 'D-']:
            reaching = nuVeto.nbody(
                resource_filename(
                'nuVeto','data/decay_distributions/D+_numu.npz'),
                esamp, enu, fn, l_ice)
        elif mother in ['Ds+', 'Ds-']:
            reaching = nuVeto.nbody(
                resource_filename(
                'nuVeto','data/decay_distributions/Ds_numu.npz'),
                esamp, enu, fn, l_ice)
        elif mother == 'K0L':
            reaching = nuVeto.nbody(
                resource_filename(
                'nuVeto','data/decay_distributions/K0L_numu.npz'),
                esamp, enu, fn, l_ice)
        else:
            # Assuming muon energy is E_parent - E_nu
            reaching = 1. - fn.prpl(list(zip((esamp-enu)*Units.GeV,
                                    [l_ice]*len(esamp))))
        return reaching

            
    @lru_cache(maxsize=2**12)
    def get_dNdEE(self, mother, daughter):
        """Differential parent-->neutrino (mother--daughter) yield"""
        ihijo = 20
        e_grid = self.mceq.e_grid
        delta = self.mceq.e_widths
        x_range = e_grid[ihijo]/e_grid
        rr = ParticleProperties.rr(mother, daughter)
        dNdEE_edge = ParticleProperties.br_2body(mother, daughter)/(1-rr)
        dN_mat = self.mceq._decays.get_matrix(
            (ParticleProperties.pdg_id[mother], 0),
            (ParticleProperties.pdg_id[daughter], 0))
        dNdEE = dN_mat[ihijo]*e_grid/delta
        logx = np.log10(x_range)
        logx_width = -np.diff(logx)[0]
        good = (logx + logx_width/2 < np.log10(1-rr)) & (x_range >= 5.e-2)

        x_low = x_range[x_range < 5e-2]
        dNdEE_low = np.array([dNdEE[good][-1]]*x_low.size)
        dNdEE_interp = lambda x_: interpolate.pchip(
            np.concatenate([[1-rr], x_range[good], x_low])[::-1],
            np.concatenate([[dNdEE_edge], dNdEE[good], dNdEE_low])[::-1],
            extrapolate=True)(x_) * np.heaviside(1-rr-x_, 1)
        return x_range, dNdEE, dNdEE_interp


    @lru_cache(maxsize=2**12)
    def grid_sol(self, ecr=None, particle=None):
        """MCEq grid solution for \\frac{dN_{CR,p}}_{dE_p}"""
        if ecr is not None:
            self.mceq.set_single_primary_particle(ecr, particle)
        else:
            self.mceq.set_primary_model(*self.pmodel)
        self.mceq.solve(int_grid=self.X_vec, grid_var="X")
        return self.mceq.grid_sol


    @lru_cache(maxsize=2**12)
    def nmu(self, ecr, particle, prpl='ice_allm97_step_1'):
        """Poisson probability of getting no muons"""
        grid_sol = self.grid_sol(ecr, particle)
        l_ice = self.geom.overburden(self.costh)
        mu = np.abs(self.get_solution('mu-', grid_sol)) + np.abs(self.get_solution('mu+', grid_sol)) # np.abs hack to prevent negative fluxes
        fn = MuonProb(prpl)
        coords = list(zip(self.mceq.e_grid*Units.GeV,
                          [l_ice]*len(self.mceq.e_grid)))
        ### DEBUG ###
        # if np.trapz(mu*fn.prpl(coords)*self.mceq.e_grid, np.log(self.mceq.e_grid)) < 0:
        #     import pdb
        #     pdb.set_trace()
        ###
        return np.trapz(mu*fn.prpl(coords)*self.mceq.e_grid, np.log(self.mceq.e_grid))


    @lru_cache(maxsize=2**12)
    def get_rescale_phi(self, mother, ecr=None, particle=None):
        """Flux of the mother at all heights"""
        grid_sol = self.grid_sol(ecr, particle) # MCEq solution (fluxes tabulated as a function of height)
        dX = self.dX_vec*Units.gr/Units.cm**2
        rho = self.mceq.density_model.X2rho(self.X_vec)*Units.gr/Units.cm**3
        inv_decay_length_array = (ParticleProperties.mass_dict[mother] / (self.mceq.e_grid[:,None] * Units.GeV)) / (ParticleProperties.lifetime_dict[mother]*rho[None,:])
        rescale_phi = dX[None,:]* inv_decay_length_array * self.get_solution(mother, grid_sol, grid_idx=False).T
        return rescale_phi


    def get_integrand(self, categ, daughter, enu, accuracy, prpl, ecr=None, particle=None):
        """flux*yield"""
        esamp = self.esamp(enu, accuracy)
        mothers = self.categ_to_mothers(categ, daughter)
        nums = np.zeros((len(esamp),len(self.X_vec)))
        dens = np.zeros((len(esamp),len(self.X_vec)))
        for mother in mothers:
            dNdEE = self.get_dNdEE(mother, daughter)[-1]
            rescale_phi = self.get_rescale_phi(mother, ecr, particle)
            # DEBUG
            # from matplotlib import pyplot as plt
            # plt.plot(np.log(self.mceq.e_grid[rescale_phi[:,0]>0]),
            #          np.log(rescale_phi[:,0][rescale_phi[:,0]>0]))
            # rescale_phi = np.array([interpolate.interp1d(self.mceq.e_grid, rescale_phi[:,i], kind='quadratic', bounds_error=False, fill_value=0)(esamp) for i in range(rescale_phi.shape[1])]).T
            ###
            # TODO: optimize to only run when esamp[0] is non-zero
            rescale_phi = np.exp(np.array([interpolate.interp1d(
                np.log(self.mceq.e_grid[rescale_phi[:,i]>0]),
                np.log(rescale_phi[:,i][rescale_phi[:,i]>0]),
                kind='quadratic', bounds_error=False, fill_value=-np.inf)(np.log(esamp))
                                           if np.count_nonzero(rescale_phi[:,i] > 0) > 2
                                           else [-np.inf,]*esamp.shape[0]
                                           for i in range(rescale_phi.shape[1])])).T
            # DEBUG
            # print rescale_phi.min(), rescale_phi.max()
            # print np.log(esamp)
            # plt.plot(np.log(esamp),
            #          np.log(rescale_phi[:,0]), label='intp')
            # plt.legend()
            # import pdb
            # pdb.set_trace()
            ###
            if 'nu_mu' in daughter:
                # muon accompanies nu_mu only
                pnmsib = self.psib(self.geom.overburden(self.costh),
                                   mother, enu, accuracy, prpl)
            else:
                pnmsib = np.ones(len(esamp))            
            dnde = dNdEE(enu/esamp)/esamp
            nums += (dnde * pnmsib)[:,None]*rescale_phi
            dens += (dnde)[:,None]*rescale_phi

        return nums, dens


    def get_solution(self,
                     particle_name,
                     grid_sol,
                     mag=0.,
                     grid_idx=None):
        """Retrieves solution of the calculation on the energy grid.

        Args:
          particle_name (str): The name of the particle such, e.g.
            ``total_mu+`` for the total flux spectrum of positive muons or
            ``pr_antinumu`` for the flux spectrum of prompt anti muon neutrinos
          mag (float, optional): 'magnification factor': the solution is
            multiplied by ``sol`` :math:`= \\Phi \\cdot E^{mag}`
          grid_idx (int, optional): if the integrator has been configured to save
            intermediate solutions on a depth grid, then ``grid_idx`` specifies
            the index of the depth grid for which the solution is retrieved. If
            not specified the flux at the surface is returned
          integrate (bool, optional): return averge particle number instead of
          flux (multiply by bin width)

        Returns:
          (numpy.array): flux of particles on energy grid :attr:`e_grid`
        """

        # MCEq index conversion
        ref = self.mceq.pman.pname2pref
        p_pdg = ParticleProperties.pdg_id[particle_name]
        reduce_res = True

        if grid_idx is None: # Surface only case
            sol = np.array([grid_sol[-1]])
            xv = np.array([self.X_vec[-1]])
        elif isinstance(grid_idx, bool) and not grid_idx: # Whole solution case
            sol = np.asarray(grid_sol)
            xv = np.asarray(self.X_vec)
            reduce_res = False
        elif grid_idx >= len(self.mceq.grid_sol): # Surface only case
            sol = np.array([grid_sol[-1]])
            xv = np.array([self.X_vec[-1]])
        else: # Particular height case
            sol = np.array([grid_sol[grid_idx]])
            xv = np.array([self.X_vec[grid_idx]])

        # MCEq solution for particle
        direct = sol[:,ref[particle_name].lidx:
                     ref[particle_name].uidx]
        res = np.zeros(direct.shape)
        rho_air = 1./self.mceq.density_model.r_X2rho(xv)

        # meson decay length
        decayl = ((self.mceq.e_grid * Units.GeV)
                  / ParticleProperties.mass_dict[particle_name]
                  * ParticleProperties.lifetime_dict[particle_name]
                  / Units.cm)

        # number of targets per cm2
        ndens = rho_air*Units.Na/Units.mol_air
        sec = self.mceq.pman[p_pdg]
        prim2mceq = {'p+-bar':'pbar-',
                     'n0-bar':'nbar0',
                     'D0-bar':'Dbar0',
                     'Lambda0-bar':'Lambdabar0'}
        for prim in self.projectiles():
            if prim in prim2mceq:
                _ = prim2mceq[prim]
            else:
                _ = prim
            prim_flux = sol[:,ref[_].lidx:
                            ref[_].uidx]
            proj = self.mceq.pman[ParticleProperties.pdg_id[prim]]
            prim_xs = proj.inel_cross_section()
            try:
                int_yields = proj.hadr_yields[sec]
                res += np.sum(int_yields[None,:,:]*prim_flux[:,None,:]*prim_xs[None,None,:]*ndens[:,None,None], axis=2)
            except KeyError as e:
                continue

        res *= decayl[None,:]
        # combine with direct
        res[direct != 0] = direct[direct != 0]

        if particle_name[:-1] == 'mu':
            for _ in ['k_'+particle_name, 'pi_'+particle_name]:
                res += sol[:,ref[_+'_l'].lidx:
                           ref[_+'_l'].uidx]
                res += sol[:,ref[_+'_r'].lidx:
                           ref[_+'_r'].uidx]

        res *= self.mceq.e_grid[None,:] ** mag

        if reduce_res:
            res = res[0]
        return res


    def get_fluxes(self, enu, kind='conv nu_mu', accuracy=3.5, prpl='ice_allm97_step_1', corr_only=False):
        """Returns the flux and passing fraction
        for a particular neutrino energy, flux, and p_light
        """
        # prpl = probability of reaching * probability of light
        # prpl -> None ==> median for muon reaching
        categ, daughter = kind.split()

        esamp = self.esamp(enu, accuracy)

        # Correlated only (no need for the unified calculation here) [really just for testing]
        passed = 0
        total = 0
        if corr_only:
            # sum performs the dX integral
            nums, dens = self.get_integrand(categ, daughter, enu, accuracy, prpl)
            num = np.sum(nums, axis=1)
            den = np.sum(dens, axis=1)
            passed = integrate.trapz(num, esamp)
            total = integrate.trapz(den, esamp)
            return passed, total
                
        pmodel = self.pmodel[0](self.pmodel[1])

        #loop over primary particles
        for particle in pmodel.nucleus_ids:
            # A continuous input energy range is allowed between
            # :math:`50*A~ \\text{GeV} < E_\\text{nucleus} < 10^{10}*A \\text{GeV}`.

            # ecrs --> Energy of cosmic ray primaries
            # amu --> atomic mass of primary

            # evaluation points in E_CR
            ecrs = amu(particle)*np.logspace(2, 10, int(10*accuracy))

            # pnm --> probability of no muon (just a poisson probability)
            nmu = [self.nmu(ecr, particle, prpl) for ecr in ecrs]

            # nmufn --> fine grid interpolation of pnm
            nmufn = interpolate.interp1d(ecrs, nmu, kind='linear',
                                         assume_sorted=True, bounds_error=False,
                                         fill_value=(0,np.nan))
            # nums --> numerator
            nums = []
            # dens --> denominator
            dens = []
            # istart --> integration starting point, the lowest energy index for the integral
            istart = max(0, np.argmax(ecrs > enu) - 1)
            for ecr in ecrs[istart:]: # integral in primary energy (E_CR)
                # cr_flux --> cosmic ray flux
                # phim2 --> units of flux * m^2 (look it up in the units)
                cr_flux = pmodel.nucleus_flux(particle, ecr.item())*Units.phim2
                # poisson exp(-Nmu) [last term in eq 12]
                pnmarr = np.exp(-nmufn(ecr-esamp))

                num_ecr = 0 # single entry in nums
                den_ecr = 0 # single entry in dens

                # dEp
                # integral in Ep
                nums_ecr, dens_ecr = self.get_integrand(categ, daughter, enu, accuracy, prpl, ecr, particle)
                num_ecr = integrate.trapz(np.sum(nums_ecr, axis=1)*pnmarr, esamp)
                den_ecr = integrate.trapz(np.sum(dens_ecr, axis=1), esamp)

                nums.append(num_ecr*cr_flux/Units.phicm2)
                dens.append(den_ecr*cr_flux/Units.phicm2)
            # dEcr
            passed += integrate.trapz(nums, ecrs[istart:])
            total += integrate.trapz(dens, ecrs[istart:])

        return passed, total


@lru_cache(maxsize=2**12)
def builder(cos_theta, pmodel, hadr, barr_mods, depth, density):
    return nuVeto(cos_theta, pmodel, hadr, barr_mods, depth, density)


def passing(enu, cos_theta, kind='conv nu_mu', pmodel=(pm.HillasGaisser2012, 'H3a'), hadr='SIBYLL2.3c', barr_mods=(), depth=1950*Units.m, density=('CORSIKA', ('SouthPole', 'December')), accuracy=3.5, fraction=True, prpl='ice_allm97_step_1', corr_only=False):
    sv = builder(cos_theta, pmodel, hadr, barr_mods, depth, density)
    num, den = sv.get_fluxes(enu, kind, accuracy, prpl, corr_only)
    return num/den if fraction else num


def fluxes(enu, cos_theta, kind='conv nu_mu', pmodel=(pm.HillasGaisser2012, 'H3a'), hadr='SIBYLL2.3c', barr_mods=(), depth=1950*Units.m, density=('CORSIKA', ('SouthPole', 'December')), accuracy=3.5, prpl='ice_allm97_step_1', corr_only=False):
    sv = builder(cos_theta, pmodel, hadr, barr_mods, depth, density)
    return sv.get_fluxes(enu, kind, accuracy, prpl, corr_only)



from setuptools import setup, find_packages


setup(
    name='nuVeto',
    version='2.1.3',
    maintainer='Tianlu Yuan',
    maintainer_email='tyuan@icecube.wisc.edu',
    description='Package implements the formalism for calculating passing fraction as discussed in JCAP07(2018)047.',
    long_description='This package calculates the effect of a detector veto on the high-energy atmospheric neutrino flux via detection of muons that reach the detector. The result calculated is the passing-flux or passing-fraction of atmospheric neutrinos as a function of energy and zenith angle.',
    url='https://github.com/tianluyuan/nuVeto.git',
    packages=find_packages('./'),
    package_data={
        'nuVeto':['data/decay_distributions/*.npz','data/prpl/*.pkl','data/corsika/*.pkl'],
        'nuVeto.resources.mu':['mmc/ice*.pklz']
    },
    install_requires=['numpy', 'scipy', 'MCEq[MKL]'],
    extras_require={
        'plotting':  ['matplotlib'],
        'resources':  ['pythia8', 'matplotlib', 'argparse', 'pandas'],
        'testing': ['pytest==4.6.4']
    },
    python_requires='>=3.3',
    license=open('LICENSE').readline().split()[0],
    classifiers=[
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering',
        ],
    )



import pickle
import os
from pkg_resources import resource_filename
from nuVeto.external import helper as exthp
from nuVeto.external import selfveto as extsv
from nuVeto.nuveto import nuVeto, passing, fluxes
from nuVeto.utils import Units, ParticleProperties, Geometry, amu, centers, calc_bins, mceq_categ_format
from nuVeto.uncertainties import BARR
from matplotlib import pyplot as plt
from matplotlib import colors
from matplotlib.ticker import LogLocator
from scipy import interpolate
import numpy as np
import crflux.models as pm


def tex(inp):
    if isinstance(inp, str):
        categ, daughter = inp.split()
        daughter_trans = {'nu_mu':r'$\nu_\mu$',
                          'nu_e':r'$\nu_e$',
                          'nu_mubar':r'$\overline{\nu}_\mu$',
                          'nu_ebar':r'$\overline{\nu}_e$'}
        return r'{} {}'.format(categ, daughter_trans[daughter])
    else:
        if inp > 1:
            return r'$E_\nu={:.2g}$'.format(inp)
        else:
            return r'$\cos \theta_z={:.2g}$'.format(inp)


def extlabel(corr_only):
    if corr_only:
        return r'$\mathcal{P}_{\rm pass}^{\rm cor, SGRS}$'
    else:
        return r'$\mathcal{P}_{\rm pass}^{\rm cor, SGRS} \mathcal{P}_{\rm pass}^{\rm uncor, GJKvS}$'        
        
    
# passing fraction tests
def fn(slice_val):
    """ decide which fn to run depending on slice_val
    """
    return pr_enu if slice_val <=1 else pr_cth


def pr_enu(cos_theta=1., kind='conv nu_mu', pmodel=(pm.HillasGaisser2012, 'H3a'), hadr='SIBYLL2.3c', barr_mods=(), depth=1950*Units.m, density=('CORSIKA', ('SouthPole', 'December')), accuracy=3.5, fraction=True, prpl='ice_allm97_step_1', corr_only=False, **kwargs):
    """ plot the passing rate (flux or fraction)
    """
    ens = np.logspace(3,7,100) if corr_only else np.logspace(3,7,39)
    passed = [passing(en, cos_theta, kind, pmodel, hadr, barr_mods, depth, density, accuracy, fraction, prpl, corr_only) for en in ens]
    if fraction:
        passed_fn = interpolate.interp1d(ens, passed, kind='quadratic')
    else:
        passed_fn = lambda es: 10**interpolate.interp1d(ens, np.log10(passed), kind='quadratic')(es)
    ens_plot = np.logspace(3,7,100)
    if fraction:
        prs = plt.plot(ens_plot, passed_fn(ens_plot), **kwargs)
        plt.ylim(0., 1.)
        plt.ylabel(r'Passing fraction')
    else:
        prs = plt.plot(ens_plot, passed_fn(ens_plot)*ens_plot**3, **kwargs)
        plt.yscale('log')
        plt.ylabel(r'$E_\nu^3 \Phi_\nu [GeV^2 cm^{-2} s^{-1} st^{-1}]$')
    plt.xlim(10**3, 10**7)
    plt.xscale('log')
    plt.xlabel(r'$E_\nu$ [GeV]')
    return prs[0]


def pr_cth(enu=1e5, kind='conv nu_mu', pmodel=(pm.HillasGaisser2012, 'H3a'), hadr='SIBYLL2.3c', barr_mods=(), depth=1950*Units.m, density=('CORSIKA', ('SouthPole', 'December')), accuracy=3.5, fraction=True, prpl='ice_allm97_step_1', corr_only=False, **kwargs):
    """ plot the passing rate (flux or fraction)
    """
    cths = np.linspace(0,1,21)
    passed = [passing(enu, cos_theta, kind, pmodel, hadr, barr_mods, depth, density, accuracy, fraction, prpl, corr_only) for cos_theta in cths]
    if fraction:
        prs = plt.plot(cths, passed, **kwargs)
        plt.ylim(0., 1.)
        plt.ylabel(r'Passing fraction')
    else:
        prs = plt.plot(cths, np.asarray(passed)*enu**3, **kwargs)
        plt.yscale('log')
        plt.ylabel(r'$E_\nu^3 \Phi_\nu [GeV^2 cm^-2 s^-1 st^-1]$')
    plt.xlim(0, 1)
    plt.xscale('linear')
    plt.xlabel(r'$\cos \theta$')
    return prs[0]


def depth(slice_val=1., kind='conv nu_mu', pmodel=(pm.HillasGaisser2012, 'H3a'), hadr='SIBYLL2.3c', fraction=True):
    depths = np.asarray([1450, 1950, 2450], 'f')*Units.m
    for depth in depths:
        fn(slice_val)(slice_val, kind, pmodel, hadr,
                      depth=depth, fraction=fraction,
                      label='depth {:.0f} m'.format(depth/Units.m))
    plt.title('{} {} {}'.format(hadr, tex(kind), tex(slice_val)))
    plt.legend()
        

def brackets(slice_val=1., kind='conv nu_mu', pmodel=(pm.HillasGaisser2012, 'H3a'), hadr='SIBYLL2.3c', fraction=True, params='g h1 h2 i w6 y1 y2 z ch_a ch_b ch_e'):
    params = params.split(' ')
    uppers = [BARR[param].error for param in params]
    lowers = [-BARR[param].error for param in params]
    all_barr_mods = [tuple(zip(params, uppers)), tuple(zip(params, lowers))]
    pr = fn(slice_val)(slice_val, kind, pmodel, hadr, label='{} {}'.format(tex(kind), tex(slice_val)))
    for barr_mods in all_barr_mods:
        fn(slice_val)(slice_val, kind, pmodel, hadr, barr_mods, fraction=fraction,
                      color=pr.get_color(), alpha=1-abs(barr_mods[0][-1]))


def samples(slice_val=1, kind='nu_mu', pmodel=(pm.HillasGaisser2012, 'H3a'), hadr='SIBYLL2.3c', fraction=True,
                 seed=88, nsamples=10, params='g h1 h2 i w6 y1 y2 z ch_a ch_b ch_e'):
    params = params.split(' ')
    pr = fn(slice_val)(slice_val, kind, pmodel, hadr=hadr, label='{} {}'.format(tex(kind), tex(slice_val)))
    np.random.seed(seed)
    for i in range(nsamples-1):
        # max(-1, throw) prevents throws that dip below -100%
        errors = [max(-1, np.random.normal(scale=BARR[param].error)) for param in params]
        barr_mods = tuple(zip(params, errors))
        fn(slice_val)(slice_val, kind, pmodel, hadr, barr_mods, color=pr.get_color(),
                      alpha=1-min(np.mean(np.abs(errors)), 0.9))


def accuracy(slice_val=1., kind='conv nu_mu', pmodel=(pm.HillasGaisser2012, 'H3a'), hadr='SIBYLL2.3c', fraction=True):
    plt.clf()
    accuracies = [2,3,4]
    for accuracy in accuracies:
        fn(slice_val)(slice_val, kind, pmodel=pmodel, hadr=hadr,
                      accuracy=accuracy, fraction=fraction,
                      label='accuracy {}'.format(accuracy))
    plt.title('{} {} {}'.format(hadr, tex(kind), tex(slice_val)))
    plt.legend()


def prpls(slice_val=1., kind='conv nu_mu', pmodel=(pm.HillasGaisser2012, 'H3a'), hadr='SIBYLL2.3c', compare=(None, 'ice_allm97_step_1', 'sigmoid_0.75_0.1')):
    for prpl in compare:
        fn(slice_val)(slice_val, kind, pmodel=pmodel, hadr=hadr, prpl=prpl,
                      label='{} {} {}'.format(prpl, tex(kind), tex(slice_val)))
    plt.legend()


def elbert(slice_val=1., kind='conv nu_mu', pmodel=(pm.GaisserHonda, None), prpl='ice_allm97_step_1', corr_only=False):
    hadrs=['DPMJET-III-3.0.6', 'SIBYLL2.3c']
    echoice = exthp.corr if corr_only else exthp.passrates
    if slice_val > 1:
        cths = np.linspace(0,1, 100)
        emu = extsv.minimum_muon_energy(extsv.overburden(cths))
        plt.plot(cths, echoice(kind)(slice_val, emu, cths), 'k--',
                 label='Analytic approx. {} {}'.format(tex(kind), tex(slice_val)))
    else:
        ens = np.logspace(2,9, 100)
        emu = extsv.minimum_muon_energy(extsv.overburden(slice_val))
        plt.plot(ens, echoice(kind)(ens, emu, slice_val), 'k--',
                 label='Analytic approx. {} {}'.format(tex(kind), tex(slice_val)))

    for hadr in hadrs:
        fn(slice_val)(slice_val, kind, pmodel, hadr, prpl=prpl, corr_only=corr_only,
                      label='{} {} {}'.format(hadr, tex(kind), tex(slice_val)))
    plt.legend()
    plt.tight_layout(0.3)


def elbert_pmodels(slice_val=1., kind='conv nu_mu', hadr='DPMJET-III-3.0.6', prpl='ice_allm97_step_1', corr_only=False):
    pmodels = [(pm.HillasGaisser2012, 'H3a', 'H3a'),
               (pm.PolyGonato, False, 'poly-gonato'),
               (pm.GaisserHonda, None, 'GH')]
    echoice = exthp.corr if corr_only else exthp.passrates
    label = extlabel(corr_only)+' {} {}'.format(tex(kind), tex(slice_val))
    if slice_val > 1:
        cths = np.linspace(0,1, 100)
        emu = extsv.minimum_muon_energy(extsv.overburden(cths))
        plt.plot(cths, echoice(kind)(slice_val, emu, cths), 'k--',
                 label=label)
    else:
        ens = np.logspace(2,9, 100)
        emu = extsv.minimum_muon_energy(extsv.overburden(slice_val))
        plt.plot(ens, echoice(kind)(ens, emu, slice_val), 'k--',
                 label=label)
    for pmodel in pmodels:
        pr = fn(slice_val)(slice_val, kind, pmodel[:2], hadr, prpl=prpl, corr_only=corr_only,
                     label='{} {} {}'.format(pmodel[2], tex(kind), tex(slice_val)))
    plt.legend()
    plt.tight_layout(0.3)


def pmodels(slice_val=1., kind='conv nu_mu', hadr='SIBYLL2.3c', prpl='ice_allm97_step_1', fraction=True):
    pmodels = [(pm.HillasGaisser2012, 'H3a', 'H3a'),
               (pm.PolyGonato, False, 'poly-gonato'),
               (pm.GaisserHonda, None, 'GH'),
               (pm.ZatsepinSokolskaya, 'default', 'ZS')]
    for pmodel in pmodels:
        pr = fn(slice_val)(slice_val, kind, pmodel[:2], hadr, prpl=prpl, fraction=fraction,
                           label='{} {} {}'.format(pmodel[2], tex(kind), tex(slice_val)))
    plt.legend()


def density_models(slice_val=1., kind='conv nu_mu', hadr='SIBYLL2.3c', prpl='ice_allm97_step_1', fraction=True):
    models = [('CORSIKA', ('BK_USStd', None)),
              ('CORSIKA', ('SouthPole', 'December')),
              ('MSIS00_IC',('SouthPole', 'June')),
              ('MSIS00_IC',('SouthPole', 'January'))]
    for model in models:
        pr = fn(slice_val)(slice_val, kind, density=model, hadr=hadr, prpl=prpl, fraction=fraction,
                           label='{} {} {} {} {}'.format(model[0], model[1][0], model[1][1], tex(kind), tex(slice_val)))
    plt.legend()


def corsika(cos_theta_bin=-1, kind='conv nu_mu', pmodel=(pm.HillasGaisser2012, 'H3a'), hadr='SIBYLL2.3', density=('CORSIKA',('SouthPole', 'December')), prpl='ice_allm97_step_1', corsika_file='eff_maxmu', plot_nuveto_lines = False, plot_legacy_veto_lines = False):
    if isinstance(cos_theta_bin, list):
        [corsika(cth, kind, pmodel, hadr, density, prpl, corsika_file, plot_nuveto_lines, plot_legacy_veto_lines) for cth in cos_theta_bin]
        return

    cfile = pickle.load(open(resource_filename('nuVeto', os.path.join('/data/corsika', corsika_file+'.pkl')), 'rb'), encoding='latin1')
    fraction = 'eff' in corsika_file
    eff, elow, eup, xedges, yedges = cfile[mceq_categ_format(kind)]
    cos_theta = centers(yedges)[cos_theta_bin]
    clean = eff[:,cos_theta_bin]>0
    xcenters = 10**centers(xedges)[clean]

    pr = plt.errorbar(xcenters, eff[:,cos_theta_bin][clean],
                     xerr=np.asarray(list(zip(xcenters-10**xedges[:-1][clean],
                                         10**xedges[1:][clean]-xcenters))).T,
                     yerr=np.asarray(list(zip(elow[:,cos_theta_bin][clean],
                                         eup[:,cos_theta_bin][clean]))).T,
                     label='CORSIKA {} {}'.format(tex(kind), tex(cos_theta)),
                     fmt='.')
    if plot_legacy_veto_lines and fraction:
        ens = np.logspace(2,9, 100)
        emu = extsv.minimum_muon_energy(extsv.overburden(cos_theta))
        plt.plot(ens, exthp.passrates(kind)(ens, emu, cos_theta), 'k--',
                 label='Analytic approx. {} {}'.format(tex(kind), tex(cos_theta)))
    if plot_nuveto_lines:
        pr_enu(cos_theta, kind, pmodel=pmodel, hadr=hadr, prpl=prpl, density=density,
               fraction=fraction, label='{} {} {}'.format(hadr, tex(kind), tex(cos_theta)), color=pr[0].get_color())
    plt.legend()


# intermediate tests
def dndee(mother, daughter):
    sv = nuVeto(0)
    x_range, dNdEE, dNdEE_interp = sv.get_dNdEE(mother, daughter)

    # print x_range[0], x_range[-1]
    x_samp = np.logspace(1, -9, 5000)
    c = plt.plot(x_samp, dNdEE_interp(x_samp), label = "Interpolated {} to {}".format(mother, daughter))
    plt.plot(x_range,dNdEE, '.', color=c[0].get_color(), label = "MCEq {} to {}".format(mother, daughter))
    plt.semilogx()
    plt.xlabel(r"$x=E_\nu/E_p$")
    plt.ylabel(r"$ \frac{dN}{dE_\nu} E_p$")
    # plt.ylim(-0.1, 5.1)
    plt.axvline(1-ParticleProperties.rr(mother, daughter), linestyle='--', color=c[0].get_color())
    plt.legend()


def plot_prpl(interp_pkl, include_mean=False, include_cbar=True):
    depth = 1950*Units.m
    prplfn = pickle.load(open(interp_pkl, 'rb'), encoding='latin1')
    emui_edges = np.logspace(2, 8, 101)
    l_ice_edges = np.linspace(1e3, 4e4, 101)
    emui = centers(emui_edges)
    l_ice = centers(l_ice_edges)
    xx, yy = np.meshgrid(emui, l_ice)
    prpls = prplfn(list(zip(xx.flatten(), yy.flatten())))
    plt.figure()
    plt.pcolormesh(emui_edges, l_ice_edges/1e3, prpls.reshape(xx.shape), cmap='magma')
    if include_cbar:
        plt.colorbar()
    if include_mean:
        small_ice = l_ice[l_ice<2.7e4]
        plt.plot(extsv.minimum_muon_energy(small_ice), small_ice/1e3, 'w--',
                 label=r'$l_{\rm ice,\,median} (E_\mu^{\rm i}, E_\mu^{\rm th} = 1\,{\rm TeV})$')
        leg = plt.legend(frameon=False, prop={'weight':'bold'}, loc='upper left')
        for text in leg.get_texts():
            plt.setp(text, color = 'w', fontsize='medium')
    plt.xlabel(r'$E_\mu^{\rm i}$ [GeV]')
    plt.ylabel(r'$l_{\rm ice}$ [km]')
    plt.locator_params(axis='y', nticks=8)
    # # plt.yscale('log')
    # # plt.gca().yaxis.set_major_formatter(ScalarFormatter())
    # plt.ticklabel_format(style='plain', axis='y')
    plt.gca().minorticks_off()
    plt.ylim(depth/Units.km, 40)
    # right y-axis with angles
    axr = plt.gca().twinx()
    axr.grid(False)
    geom = Geometry(depth)
    costhetas = geom.overburden_to_cos_theta(np.arange(10, 41, 10)*Units.km)
    axr.set_ylim(depth/Units.km, 40)
    axr.set_yticks(geom.overburden(costhetas)/1e3)
    axr.set_yticklabels(np.round(costhetas, 2))
    axr.set_ylabel(r'$\cos \theta_z$', rotation=-90)
    axr.set_xscale('log')
    axr.set_xlim(1e2, 1e8)
    axr.minorticks_off()
    xlocmaj = LogLocator(base=10,numticks=12)
    axr.get_xaxis().set_major_locator(xlocmaj)
    return emui_edges, l_ice_edges, prpls.reshape(xx.shape)


def plot_prpl_ratio(interp_pkl_num, interp_pkl_den, include_cbar=True):
    emui_edges, l_ice_edges, prpls_num = plot_prpl(interp_pkl_num, False, False)
    emui_edges, l_ice_edges, prpls_den = plot_prpl(interp_pkl_den, False, False)
    plt.figure()
    plt.pcolormesh(emui_edges, l_ice_edges/1e3, np.ma.masked_invalid(prpls_num/prpls_den),
                   norm=colors.LogNorm(vmin=1e-2, vmax=1e2),
                   cmap='coolwarm')
    if include_cbar:
        plt.colorbar()
    plt.xlabel(r'$E_\mu^{\rm i}$ [GeV]')
    plt.ylabel(r'$l_{\rm ice}$ [km]')
    # plt.yscale('log')
    # plt.gca().yaxis.set_major_formatter(ScalarFormatter())
    plt.xscale('log')
    xlocmaj = LogLocator(base=10,numticks=12)
    plt.gca().xaxis.set_major_locator(xlocmaj)
    plt.gca().minorticks_off()
    plt.ticklabel_format(style='plain', axis='y')
    plt.xlim(1e2, 1e8)
    plt.ylim(1, 40)
    plt.title('{}/{}'.format(os.path.splitext(os.path.basename(interp_pkl_num))[0],
                             os.path.splitext(os.path.basename(interp_pkl_den))[0]))

    
def parent_ratio(cos_theta, parents='pi+ pi-', pmodel=(pm.HillasGaisser2012, 'H3a'), hadr='SIBYLL2.3c',
                 ecr=None, particle=None):
    plt.figure()
    sv = nuVeto(cos_theta, pmodel, hadr)
    gsol = sv.grid_sol(ecr, particle)
    X_vec = sv.X_vec
    for idx, x_val in enumerate(X_vec):
        calc0 = sv.get_solution(parents.split()[0], gsol, grid_idx=idx)
        calc1 = sv.get_solution(parents.split()[1], gsol, grid_idx=idx)
        plt.plot(sv.mceq.e_grid, calc0/calc1,
                 label='X={:.3g} km'.format(
                     float(sv.mceq.density_model.X2h(x_val))/1e5))

    plt.xlabel(r'$E_p$')
    plt.ylabel(r'{}/{} flux ratio'.format(*parents.split()))
    plt.xscale('log')
    plt.legend()
    if ecr is None or particle is None:
        plt.title(r'$\cos \theta = {:.2g}$'.format(cos_theta))
    else:
        plt.title(r'{} at {:.2g} GeV and $\cos \theta = {:.2g}$'.format(particle, ecr, cos_theta))


def parent_flux(cos_theta, parent='D0', pmodel=(pm.HillasGaisser2012, 'H3a'), hadr='SIBYLL2.3c',
                density=('CORSIKA',('SouthPole', 'December')), mag=3, ecr=None, particle=None):
    plt.figure()
    sv = nuVeto(cos_theta, pmodel, hadr, density=density)
    gsol = sv.grid_sol(ecr, particle)
    ### hack to convert list to np array
    sv.mceq.grid_sol = np.asarray(sv.mceq.grid_sol)
    ###
    X_vec = sv.X_vec
    for idx, x_val in enumerate(X_vec):
        mceq = sv.mceq.get_solution(parent, mag, grid_idx=idx)
        mceq[mceq==0] = np.nan
        calc = sv.get_solution(parent, gsol, mag, grid_idx=idx)
        pout = plt.loglog(sv.mceq.e_grid, mceq,
                          label='$X={:.3g}$ km'.format(
                              float(sv.mceq.density_model.X2h(x_val))/1e5))
        plt.loglog(sv.mceq.e_grid, calc, '--',
                   color=pout[0].get_color())

    plt.xlabel(r'$E_p$')
    plt.ylabel(r'$E_p^{} \Phi_p$'.format(mag))
    plt.ylim(ymin=1e-20)
    plt.legend()
    plt.title('{} {}'.format(parent, tex(cos_theta)))
    plt.tight_layout(0.3)
    # plt.savefig('/Users/tianlu/Desktop/selfveto/parent_flux/combined/{}.png'.format(parent))
        

def nu_flux(cos_theta, kinds='conv nu_mu', pmodel=(pm.HillasGaisser2012, 'H3a'), hadr='SIBYLL2.3c', mag=3, logxlim=(3,7), corr_only=False):
    sv = nuVeto(cos_theta, pmodel, hadr)
    sv.grid_sol()
    fig, axs = plt.subplots(2,1)
    for kind in [_.strip() for _ in kinds.split(',')]:
        plt.sca(axs[0])
        mine = np.asarray([fluxes(en, cos_theta, kind, pmodel, hadr, corr_only=corr_only)[1] for en in sv.mceq.e_grid])
        pr = plt.plot(sv.mceq.e_grid, mine*sv.mceq.e_grid**mag,
                      label='{} {} {}'.format(hadr, tex(kind), tex(cos_theta)))
        plt.ylabel(r'$E_\nu^{} \Phi_\nu$'.format(mag))
        plt.loglog()
        plt.xlim(*np.power(10,logxlim))
        plt.ylim(ymin=1e-8)
        plt.legend()

        try:
            theirs = sv.mceq.get_solution(mceq_categ_format(kind))
            pr = plt.plot(sv.mceq.e_grid, theirs*sv.mceq.e_grid**mag,
                          linestyle='--', color=pr[0].get_color())

            plt.sca(axs[1])
            plt.plot(sv.mceq.e_grid, theirs/mine)
        except KeyError:
            plt.sca(axs[1])
        finally:
            plt.ylabel(r'ratio MCEq/Calc')
            plt.xscale('log')
            plt.ylim(0.5, 1.9)
            plt.xlabel(r'$E_\nu$ [GeV]')
            plt.xlim(*np.power(10,logxlim))


def prob_nomu(cos_theta, particle=14, pmodel=(pm.HillasGaisser2012, 'H3a'), hadr='SIBYLL2.3c', prpl='ice_allm97_step_1'):
    """ plot prob_nomu as fn of ecr
    """
    ecrs = amu(particle)*np.logspace(3, 10, 20)
    ecrs_fine = amu(particle)*np.logspace(3, 10, 1000)
    sv = nuVeto(cos_theta, pmodel, hadr)
    nmu = [sv.nmu(ecr, particle, prpl) for ecr in ecrs]
    nmufn = interpolate.interp1d(ecrs, nmu, kind='linear',
                                 assume_sorted=True, fill_value=(0,np.nan))
    plt.semilogx(ecrs_fine, np.exp(-nmufn(ecrs_fine)), label='interpolated')
    plt.semilogx(ecrs, np.exp(-np.asarray(nmu)), 'ko')
    plt.xlabel(r'$E_{CR} [GeV]$')
    plt.ylabel(r'$e^{-N_\mu}$')
    plt.legend()


def elbert_only(slice_val=1., kind='conv nu_mu'):
    plt.figure()
    if 'nu_e' in kind:
        echoices = [exthp.passrates]
        names = [r'$\mathcal{P}_{\rm pass}^{\rm uncor, GJKvS}$']
    else:
        echoices = [exthp.corr, exthp.passrates]
        names = [extlabel(True),
                 extlabel(False)]
    if slice_val > 1:
        cths = np.linspace(0,1, 100)
        emu = extsv.minimum_muon_energy(extsv.overburden(cths))
        for echoice, name in zip(echoices,names):
            plt.plot(cths, echoice(kind)(slice_val, emu, cths),
                     '--', label=name)
    else:
        ens = np.logspace(2,9, 100)
        emu = extsv.minimum_muon_energy(extsv.overburden(slice_val))
        for echoice, name in zip(echoices,names):
            plt.plot(ens, echoice(kind)(ens, emu, slice_val), '--',
                     label=name)
    plt.title(r'{}, {}'.format(tex(kind), tex(slice_val)))
    plt.ylim(0., 1.)
    plt.ylabel(r'Passing fraction')
    plt.xlim(10**3, 10**7)
    plt.xscale('log')
    plt.xlabel(r'$E_\nu$ [GeV]')
    plt.legend()
    plt.tight_layout(0.3)


def hist_preach(infile, plotdir=None):
    import pandas as pd
    napf = 36
    df = pd.read_csv(infile, delim_whitespace=True, header=None,
                     names='ei l ef'.split())
    # If the muon doesn't reach, MMC saves ef as -distance traveled
    df[df<0] = 0
    for idx, (ei, sdf) in enumerate(df.groupby('ei')):
        if idx % napf == 0:
            if idx > 0:
                # plt.legend(fontsize=6)
                plt.tight_layout()
                if plotdir is not None:
                    plt.savefig(os.path.join(
                        os.path.expanduser(plotdir), '{}.png'.format((idx-1)/napf)))
            fig, axs = plt.subplots(6, 6, figsize=(10,10))
            # fig.text(0.5, 0.04, r'$E_f$', ha='center', va='center')
            # fig.text(0.06, 0.5, r'$P(E_f|E_i, l)$', ha='center', va='center', rotation='vertical')
            axs = axs.flatten()
        ax = axs[idx%napf]
        ax.set_prop_cycle('color',plt.cm.Blues(np.linspace(0.3,1,100)))
        plt.sca(ax)
        plt.title(r'${:.2g}$ GeV'.format(ei), fontdict={'fontsize':8})
        for l, efs in sdf.groupby('l'):
            bins = calc_bins(efs['ef'])
            counts, edges = np.histogram(efs['ef'], bins=bins, density=True)
            plt.plot(centers(edges), counts, label='{:.3g} km'.format(l/1e3))
        plt.yscale('log')
        # plt.xlim(sdf['ef'].min(), sdf['ef'].max()*1.1)
    # plt.legend(fontsize=6)
    plt.tight_layout()
    if plotdir is not None:
        plt.savefig(os.path.join(
            os.path.expanduser(plotdir), '{}.png'.format((idx-1)/napf)))



import os
import pickle
from pkg_resources import resource_filename
from MCEq.geometry.geometry import EarthGeometry
import mceq_config as config
import numpy as np
from scipy import stats
from particletools.tables import SibyllParticleTable, PYTHIAParticleData


class Units(object):
    # units
    Na = 6.0221415e+23 # mol/cm^3
    km = 5.0677309374099995 # km to GeV^-1 value from SQuIDS
    cm = km*1.e-5
    m = km*1.e-3
    gr = 5.62e23 # gr to GeV value from SQuIDS
    sec = 1523000.0 #$ sec to GeV^-1 from SQuIDS
    day = sec*60**2*24
    yr = day*356.25
    GeV = 1
    MeV = 1e-3*GeV
    TeV = 1.e3*GeV
    PeV = 1.e3*TeV
    mol_air = config.A_target
    phim2 = (m**2*GeV*sec)**-1
    phicm2 = (cm**2*GeV*sec)**-1


class ParticleProperties(object):
    modtab = SibyllParticleTable()
    pd = PYTHIAParticleData()
    
    mass_dict = {}; lifetime_dict = {}; pdg_id = {}; sibling = {};

    for k in modtab.part_table:
        pdg_id[k] = modtab.modname2pdg[k]
        mass_dict[k] = pd.mass(pdg_id[k]) * Units.GeV
        lifetime_dict[k] = pd.ctau(pdg_id[k]) * Units.cm

    @staticmethod
    def rr(mother, daughter):
        """ returns ratio of masses
        """
        other_masses = []
        mother_pdg = ParticleProperties.pdg_id[mother]
        daughter_pdg = ParticleProperties.pdg_id[daughter]
        for br, prod in ParticleProperties.pd.decay_channels(mother_pdg):
            if daughter_pdg in prod:
                mass_tot = sum([ParticleProperties.pd.mass(abs(prod))
                                for prod in prod])-ParticleProperties.pd.mass(daughter_pdg)
                other_masses.append(mass_tot)
                
        return (min(other_masses)/ParticleProperties.mass_dict[mother])**2


    @staticmethod
    def br_2body(mother, daughter):
        """ returns the two-body branching ratio if it exists
        """
        mother_pdg = ParticleProperties.pdg_id[mother]
        daughter_pdg = ParticleProperties.pdg_id[daughter]
        brs = 0
        for br, prod in ParticleProperties.pd.decay_channels(mother_pdg):
            if daughter_pdg in prod and len(prod) == 2:
                brs += br 
        return brs


class MuonProb(object):
    def __init__(self, pklfile):
        if pklfile is None:
            self.mu_int = self.median_approx
        elif os.path.isfile(pklfile):
            self.mu_int = pickle.load(open(pklfile, 'rb'), encoding='latin1')
        else:
            self.mu_int = pickle.load(open(resource_filename('nuVeto', os.path.join('data', 'prpl', pklfile+'.pkl')), 'rb'), encoding='latin1')


    def median_emui(self, distance):
        """
        Minimum muon energy required to survive the given thickness of ice with at
        least 1 TeV 50% of the time.

        :returns: minimum muon energy [GeV] for 1 TeV
        """
        # require that the muon have median energy 1 TeV
        b, c = 2.52151, 7.13834
        return 1e3 * np.exp(1e-3 * distance / (b) + 1e-8 * (distance**2) / c)

    
    def median_approx(self, coord):
        coord = np.asarray(coord)
        muon_energy, ice_distance = coord[...,0], coord[...,1]
        min_mue = self.median_emui(ice_distance)*Units.GeV
        return muon_energy > min_mue

    
    def prpl(self, coord):
        pdets = self.mu_int(coord)
        pdets[pdets > 1] = 1
        return pdets
        
    
class Geometry(EarthGeometry):
    def __init__(self, depth):
        """ Depth of detector and elevation of surface above sea-level
        """
        super(Geometry, self).__init__()
        self.depth = depth
        self.h_obs *= Units.cm
        self.h_atm *= Units.cm
        self.r_E *= Units.cm
        self.r_top = self.r_E + self.h_atm
        self.r_obs = self.r_E + self.h_obs


    def overburden(self, cos_theta):
        """Returns the overburden for a detector at *depth* below some surface
        at *elevation*.

        From law of cosines,
        x^2 == r^2+(r-d)^2-2r(r-d)cos(gamma)
        where
        r*cos(gamma) = r-d+x*cos(theta), solve and return x.

        :param cos_theta: cosine of zenith angle in detector coord
        """
        d = self.depth
        r = self.r_E
        z = r-d
        return (np.sqrt(z**2*cos_theta**2+d*(2*r-d))-z*cos_theta)/Units.m


    def overburden_to_cos_theta(self, l):
        """Returns the theta for a given overburden for a detector at 
        *depth* below some surface at *elevation*.

        From law of cosines,
        x^2 == r^2+(r-d)^2-2r(r-d)cos(gamma)
        where
        r*cos(gamma) = r-d+x*cos(theta), solve and return x.

        :param cos_theta: cosine of zenith angle in detector coord
        """
        d = self.depth
        r = self.r_E
        z = r-d
        return (2*d*r-d**2-l**2)/(2*l*z)


    def cos_theta_eff(self, cos_theta):
        """ Returns the effective cos_theta relative the the normal at earth surface.

        :param cos_theta: cosine of zenith angle (detector)
        """
        d = self.depth
        r = self.r_E
        z = r-d
        return np.sqrt(1-(z/r)**2*(1-cos_theta**2))
        
    
def amu(particle):
    """
    :param particle: primary particle's corsika id

    :returns: the atomic mass of particle
    """
    return 1 if particle==14 else particle/100


def centers(x):
    return (x[:-1]+x[1:])*0.5


def calc_nbins(x):
    ptile = np.percentile(x, 75) - np.percentile(x, 25)
    if ptile == 0:
        return 10
    n =  (np.max(x) - np.min(x)) / (2 * len(x)**(-1./3) * ptile)
    return np.floor(n)


def calc_bins(x):
    nbins = calc_nbins(x)
    edges = np.linspace(np.min(x), np.max(x)+2, num=nbins+1)
    # force a bin for 0 efs
    if edges[0] == 0 and edges[1]>1:
        edges = np.concatenate((np.asarray([0., 1.]),edges[1:]))
    return edges


def mceq_categ_format(kind):
    _c, _d = kind.split()
    if 'bar' in _d:
        _d = 'anti'+_d.replace('bar', '')
    return _c+'_'+_d.replace('_', '')
